unit untBoardDownloadThread;

interface

uses
  Classes, SysUtils,
  RegExpr,
  untStreamTool, untTool,
  untHttpThread, untBoard, untTopic, untBBSSub,
  untGlobal;

type

  TBoardDownloadThread = class(THttpThread)
  private
    FBoard        : TOnlineBoard;
    FBuffer       : TMemoryStream;
    FBufferReader : TStreamReader;
    FLastPosition : integer;
    FCurrentIndex : integer;
    procedure HttpReceived(const Buff; Count : int64); override;
  public
    constructor Create(Board : TOnlineBoard);
    destructor  Destroy; override;
    procedure   Execute; override;
  end;

implementation

{ TBoardDownloadThread }

constructor TBoardDownloadThread.Create(Board: TOnlineBoard);
begin
  inherited Create(Board.SubjectTxtUrl);

  FBoard  := Board;
  FBuffer := TMemoryStream.Create();
  FBufferReader := TStreamReader.Create(FBuffer);
  FLastPosition := 0;
  FCurrentIndex := 0;

end;

destructor TBoardDownloadThread.Destroy;
begin
  FBuffer.Free;
  FBufferReader.Free;

  inherited;
end;

procedure TBoardDownloadThread.Execute;
begin

  // s
  inherited;

  FBoard.IsDownloading := false;

  Terminate;

end;

procedure TBoardDownloadThread.HttpReceived(const Buff; Count: int64);
var
  line          : string;
  Regex         : TRegExpr;
  topic         : TTopic;
  topicId       : string;
  topicPath     : string;
  topicTitle    : string;
  topicMsgCount : integer;
begin

  Regex := TRegExpr.Create;

  // sPʂ
  FBuffer.Seek(0, soFromEnd);
  FBuffer.Write(Buff, Count);
  FBuffer.Seek(FLastPosition, soFromBeginning);

  while FBufferReader.ReadLine(line) do
  begin
    FLastPosition := FBuffer.Position;

    case FBoard.BBSType of
      bt2ch,
      btMitinoku:  Regex.Expression := '^(.+?)\.dat<>(.*) \((.+?)\)$';
      btJBBS:      Regex.Expression := '^(.+?)\.cgi,(.*)\((.+)\)$';
      btShitaraba: Regex.Expression := '^(.+?)\.dat<>(.+?)<>(.+?)<>';
    end;

    if Regex.Exec(line) then
    begin
      Inc(FCurrentIndex);

      topicId       := Regex.Substitute('$1');
      topicTitle    := Regex.Substitute('$2');
      topicMsgCount := StrToIntNeo(Regex.Substitute('$3'));

      // ꕶϊ
      topicTitle := StringReplace(topicTitle, '&gt;', '>', [rfReplaceAll]);
      topicTitle := StringReplace(topicTitle, '&lt;', '<', [rfReplaceAll]);

      // gsbN̐
      //topicPath          := FBoard.BoardPath + '/' + topicId;
      topic              := gBBSCore.GetTopic(FBoard, topicId);
      topic.Index        := FCurrentIndex;
      topic.Title        := topicTitle;
      topic.MessageCount := topicMsgCount;
      topic.NoIndex      := false;
    end;

    // Cxgs
    FBoard.RaiseTopicReceivedEvent(topic);
  end;

  Regex.Free;

  inherited;

end;

end.
